#ifndef	DV_HTTPHEAD_H
#define	DV_HTTPHEAD_H
// $Id: httphead.h,v 1.10 2003/08/16 08:30:34 dvermeir Exp $

#include <sys/types.h>
#include <iostream>
#include <string>
#include <list>
#include <map>

namespace Dv { namespace Cgi {
/** A class representing the header part of a HTTP response.
 *
 * The class makes it convenient to set up standard headers and
 * cookies for the response:
 * <ul>
 * <li> Dv::Cgi::HttpHeader::nocache() generates header lines assuring
 *   that a browser will not keep the response pagein its cache.
 * <li> Dv::Cgi::HttpHeader::cookie() sets the a cookie 
 * <li> Dv::Cgi::HttpHeader::content_type() sets the content-type
 *   of the page.
 * <li> Dv::Cgi::HttpHeader::location() redirects the browser to a url
 * <li> Dv::Cgi::HttpHeader::refresh() redirects the browser after a
 *   certain period
 * </ul>
 */
class HttpHeader {
public:
  /** Constructor
   * @param content type, default is "text/plain"
   * @param debug debug flag
   */
  HttpHeader(const std::string& content="text/plain", bool debug=false);
  /** Type of data member that contains headers (except cookies) */
  typedef std::map<std::string,std::string> header_map;
  /** Iterator type that allows access to header strings (except
   * cookies). */
  typedef header_map::iterator  iterator;
  /** Const Iterator type that allows access to header strings (except
   * cookies). */
  typedef header_map::const_iterator  const_iterator;

  /** @return current debug status */
  bool  debug() const { return debug_; }
  /** @return const_iterator pointing to first header string */
  const_iterator begin() const { return headers_.begin(); }
  /** @return const_iterator pointing past last header string */
  const_iterator end() const { return headers_.end(); }
  /** @return iterator pointing to first header string */
  iterator begin() { return headers_.begin(); }
  /** @return iterator pointing past last header string */
  iterator end() { return headers_.end(); }
  /** @return number of header strings, excluding cookies. */
  size_t size() const { return headers_.size(); }

  /** Add header strings that prevent browser from caching the page. */
  HttpHeader& nocache();
  /** Set cookie name = value for path.
   * Providing an empty value will remove the cookie. 
   * @param name of cookie
   * @param value of cookie (default is "", i.e. delete)
   * @param path of cookie (default is "/")
   * @return reference to *this.
   */
  HttpHeader& cookie(const std::string& name, const std::string& value="", 
    const std::string& path="/");
  /** Add header strings to set content type.
   * @param value content type, e.g. "text/html"
   * @return reference to *this.
   */
  HttpHeader& content_type(const std::string& value);
  /** Add header string to redirect browser to a given url.
   * @param value url to redirect to
   * @return reference to *this.
   */
  HttpHeader& location(const std::string& value);
  /** Add header string to refresh browser
   * @param duration in seconds that browser will wait before refresh,
   *   default is 0
   * @param url which will be accessed upon refresh,
   *   default is same url
   * @return reference to *this.
   */
  HttpHeader& refresh(size_t duration=0, const std::string& url="");

  /** General header add function.
   * @param name first part of header line, e.g. "Content-type: ".
   * @param value rest of header line, e.g."text/html".
   * @return reference to *this.
   * Example
   * @code
   * httpheader.add("Content-type: ","text/html");
   * @endcode
   * @warning
   * Any previous value of "name" is replaced by the new value.
   */
  HttpHeader& add(const std::string& name, const std::string& value);
  /** Remove any header with a given name.
   * @param name first part of header to remove
   * @return reference to *this.
   * @sa Dv::Cgi::HttpHeader::add
   */
  HttpHeader& erase(const std::string& name);
  /** Remove all header strings.
   * @return reference to *this.
   */
  HttpHeader& clear() { headers_.clear(); return *this; }

  /** Output header strings and cookie definitions.
   * @param os to write headers to
   * @param header to write
   * @return reference to os
   */
  friend std::ostream& operator<<(std::ostream& os, const HttpHeader& header);
private:
  bool debug_;
  header_map headers_;	
  typedef std::map<std::string,std::pair<std::string,std::string> > cookie_map;
  cookie_map cookies_;
};
} }
#endif
