// $Id: sessionserver.C,v 1.12 2004/03/20 14:20:44 dvermeir Exp $

#include <string>
#include <unistd.h> // pid_t, fork(), pause(), getpid()
#include <stdexcept>
#include <fstream>

#include <dvutil/makedaemon.h> // make_daemon()
#include <dvutil/file.h> // make_daemon()
#include <dvutil/logstream.h> 
#include <dvutil/ntostr.h> 
#include <dvutil/date.h> 

#include <config.h>
#include "cgidata.h"
#include "formdata.h"
#include "cookieprops.h"
#include "sessionserver.h"

namespace Dv {
namespace Cgi {

/////////////////////////////////////////////////////////////////////////////
// SessionServer::Filter
/////////////////////////////////////////////////////////////////////////////
SessionServer::Filter::Filter(SessionServer& server): count_(0), 
  server_(server) {
}

int
SessionServer::Filter::get() {
return EOF;
}

int
SessionServer::Filter::put(int c) {
if (c!=EOF) {
  if (count_==0)
    *(server_.client_socket_) << server_.header();
  ++count_;
  return (server_.client_socket_->put(c) ? c : EOF);
  }
return EOF;
}         

int
SessionServer::Filter::sync() {
server_.log() << "filter.sync()" << std::endl;
return (server_.client_socket_->rdbuf()->sync() == 0 ? 0 : -1);
} 

/////////////////////////////////////////////////////////////////////////////
// SessionServer
/////////////////////////////////////////////////////////////////////////////

SessionServer::SessionServer(
  const std::string& program_name, 
  time_t delay,
  const std::string& timeouturl,
  const std::string& errorurl,
  const std::string& logdir,
  const std::string& socketdir
  ): std::ostream(0), name_(program_name), delay_(delay), timeouturl_(timeouturl), 
   server_socket_(0), client_socket_(0), status_(true), log_(0), filter_(*this) {
std::string prefix(program_name + "-client." + Util::ntostr(getpid()) );
std::string logfile(logdir + "/" + program_name + ".log" );

log_ = new Util::logstream(logfile,prefix);

if (!*log_) {
  std::cerr << Util::Date().str() << " " << prefix << ": cannot open " << logfile << std::endl;
  exit(1);
  }

*log_ << "start" << std::endl;

// Get data, if any.

Util::Props  data;
Util::Props  cookies;

std::string raw_data;
cgidata(raw_data);
formdata(raw_data,data);
cookieprops(cookies);

const std::string* ps = cookies.find(program_name);

*log_ << "cookie = " << (ps ? *ps : std::string("(none)")) << std::endl;

// Check whether the file *ps exists; if not zap ps.

if (ps) {
  Util::File fserver(*ps);
  if (fserver.exists())
    server_filename_ = *ps;
  else {
    /********************************************************************/
    /* CLIENT -- error (timeout) response				*/
    /********************************************************************/
    HttpHeader header;
    *log_ << "stale cookie redirect => " << timeouturl << std::endl;
    header.cookie(program_name).location(timeouturl);
    std::cout << header << std::endl;
    exit(1);
    }
  }

if (ps==0) {
  // First call to this program or stale cookie: (re)start server as a
  // child daemon process.

  // But first, determine the name so that also the client will know it.
  Util::Directory dir(socketdir);
  server_filename_ = dir.tempfile(program_name+".sock");
  // Since Dv::Util::Directory::tempfile() uses mkstemp(), the
  // file actually exists, which we don't want here since that would
  // let the (forked off) client think too soon that the server is up.
  // So, we remove it, hoping that the chances for another process
  // recreating it in the mean time are very small..
  Util::File f(server_filename_);
  if (f.exists())
    f.rm();

  // Fork and let child become server.
  *log_ << "forking.." << std::endl;
  pid_t	pid = fork();

  if (pid==0) {
    /********************************************************************/
    /* SERVER								*/
    /********************************************************************/

    // Child process will become server.
    delete log_;
    prefix = program_name + "-server." + Util::ntostr(getpid());
    log_ = new Util::logstream(logfile, prefix);
    *log_ << "start.." << std::endl;

    Util::make_daemon(); // Become daemon.
    server_socket_ = new Net::userversocket(server_filename_,delay);
    rdbuf(new Dv::Util::filterstreambuf<Filter>(filter_));
    *log_ << "userversocket status = " << server_socket_->strerror() << std::endl;

    get_input();

    // *log_ << "input: " << std::endl << input() << std::endl;

    header().cookie(program_name,server_filename_);

    *log_ << "initialization complete" << std::endl;
    return; // Server returns from constructor.
    }
  }

/************************************************************************/
/* CLIENT 								*/
/************************************************************************/

// Client cgi stub program: connect to server.

// Wait for server to become available.
{
int n(0);
Util::File fserver(server_filename_);
while ((!fserver.exists()) && (n++<5))
  sleep(1);
}

// Open connection to server.
Net::usocket* socket = new Net::usocket(server_filename_, 0);
*log_ << "connection to server = " << socket->strerror() << std::endl;

if (*socket) {
  /**********************************************************************/
  /* CLIENT -- response generated by server			    	*/
  /**********************************************************************/
  // Send data to server.
  // 1. the formn data are already  present in ``global scope''.
  // 2. Environment in env:: scope.
  Util::Props env;
  Util::env2props(env);
  data.merge(env,"env");
  // 3. Cookies in cookies:: scope.
  data.merge(cookies,"cookies");
  // Actual send.
  data.send(*socket);
  // Now copy input from server to cout.
  std::cout << socket->rdbuf();
  if (!*socket) 
    *log_ << "server not responding" << std::endl;
  delete socket;
  exit(0);
  }

/************************************************************************/
/* CLIENT -- error response						*/
/************************************************************************/

*log_ << "cannot contact server " << server_filename_ << std::endl;
HttpHeader  header;
delete socket;
header.cookie(program_name).location(errorurl);
std::cout << header << std::endl;
exit(2);
}

SessionServer::~SessionServer() {
*log_ << "auto-destruct" << std::endl;
if (log_)
  delete log_;
if (client_socket_)
  delete client_socket_;
if (server_socket_)
  delete server_socket_;
delete rdbuf();
}

std::string
SessionServer::here() const throw (std::runtime_error) {
return env("REQUEST_URI"); // includes possible parameters
}

std::string
SessionServer::here_base() const throw (std::runtime_error) {
std::string uri = here();
std::string::size_type qmark(uri.find('?'));
if ( qmark == std::string::npos )
  return uri;
else
  return uri.substr(0, qmark);
}

void
SessionServer::get_input() {
// Flush & close previous connection from client.
*log_ << "get input start" << std::endl;
if (client_socket_) 
  delete client_socket_;
// Wait for connection from client.
client_socket_ = new Net::usocket(*server_socket_,delay_,1024);
*log_ << "get input: socket created " << std::endl;
filter_.clear();
props_.clear();
env_.clear();
cookies_.clear();
if (status_ = bool(*client_socket_))
  props_.receive(*client_socket_);
else
  *log_ << "get input: socket bad state " << std::endl;

props_.select("cookies",cookies_);
props_.select("env",env_);

*log_ << "get_input " << client_socket_->strerror() << std::endl;
}


void
SessionServer::refresh() {
*log_ << "refresh count = " << filter_.count() << std::endl;
if (filter_.count()==0) {
  const std::string prefix("Dv::Cgi::SessionServer::refresh error");
  const std::string title(prefix + " (" + name() + ")");
  header().content_type("text/html");
  (*this) << "<html><head><title>\n" 
          << title  
	  << "</title></head>\n<body bgcolor=\"#ffffff\">\n"
          << "<h2>" << title << "</h2>\n" 
          << "<p> The refresh function for a <i>SessionServer</i>\n"
             "object was called before any output was written to it.\n"
             "<p> This is an internal error in the application's use\n"
             "of the <i>Dv::Cgi::SessionServer</i> class in the\n"
             "<a href=\"http://tinf2.vub.ac.be/~dvermeir/software/dv/index.html\">"
             "dvcgi</a> package.\n"
          << "<hr>\n"
             "<a href=\"mailto:dvermeir@vub.ac.be\">Dirk Vermeir</a>\n"
	     "</body></html>"
          << std::endl;
  }
rdbuf()->STREAMBUF_PUBSYNC();
get_input();
}

void
SessionServer::disconnect(const std::string& url) {
*log_ << "disconnect: " << url << std::endl;
header().cookie(name()).location(url).content_type("text/plain");
(*this) << std::endl;
flush();
disconnect();
}

void
SessionServer::disconnect() {
delete client_socket_; client_socket_ = 0;
delete server_socket_; server_socket_ = 0;
}

void
eof(SessionServer& s) {
s.refresh();
}

}}

void
operator<<(std::ostream& os, void (*f)(Dv::Cgi::SessionServer& s)) {
Dv::Cgi::SessionServer*	ss = dynamic_cast<Dv::Cgi::SessionServer*>(&os);
if (ss)
  f(*ss);
}
